local AddonName, GGL = ...
--[[-----------------------------------------------------------------------------
Label Widget
Displays text and optionally an icon.

Events:
OnInviteAccept(self, event, oldState, sessionID, sessionType)
OnInviteDecline(self, event, oldState, sessionID, sessionType)
-------------------------------------------------------------------------------]]
local Type, Version = "GoldGoblinLootGUIWidget-InvitationLine", 1
local AceGUI = LibStub and LibStub("AceGUI-3.0", true)
if not AceGUI or (AceGUI:GetWidgetVersion(Type) or 0) >= Version then return end

local L = GGL.Locale

-- Lua APIs
local max, select, pairs = math.max, select, pairs
local tonumber = tonumber
local abs, floor = math.abs, math.floor
local format = string.format

-- WoW APIs
local CreateFrame, UIParent = CreateFrame, UIParent
local GetItemIcon, GetItemInfo = GetItemIcon, GetItemInfo

--[[-----------------------------------------------------------------------------
Support functions
-------------------------------------------------------------------------------]]
local GAP_SIZE = 2
local TOP_HEIGHT, MIN_HEIGHT = 15, 14
local UNSELECTED_ALPHA = 0.3

local STATE_ID_LIST = GGL.INVITE_KEY_LIST
local STATE_LIST = {
    [STATE_ID_LIST.INVITED] = "groupfinder-waitdot",
    [STATE_ID_LIST.ACCEPTED] = "groupfinder-icon-greencheckmark",
    [STATE_ID_LIST.DECLINED] = "communities-icon-redx",
}

local function AcceptDecline_OnClick(self, ...)
    if self.obj.isDisabled then return end
    local typ = self.type
    if typ == self.obj.selectedType then return end
    local oldState = self.obj.selectedType
    local event
    if typ == STATE_ID_LIST.ACCEPTED then
        event = "OnInviteAccept"
    elseif typ == STATE_ID_LIST.DECLINED then
        event = "OnInviteDecline"
    end
    self.obj:SetSelectedState(typ)

    self.obj:Fire(event, oldState, self.obj.inviteTemplate)
end

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {
    ["OnAcquire"] = function(self)
        self.selectedType = nil

        self:SetSelectedState()
        self:SetSessionID()
        self:SetName()
        self:SetDescription()
        self:Disable(false)
        self:SetTypeText()
    end,

    ["OnRelease"] = function(self)
        self:UnregisterInviteTemplateEvents()
        self:SetSessionID()
    end,

    ["OnWidthSet"] = function(self, width)

    end,

    ["SetName"] = function(self, name)
        self.textName:SetText(name or "")
    end,

    ["SetDescription"] = function(self, description)
        local text = self.textDesc
        description = (not description or description=="") and L["<no description provided>"] or description
        text:SetHeight(1000)
        text:SetText(description or "")
        text:SetHeight(text:GetStringHeight())
        self.frame:SetHeight(text:GetHeight() + TOP_HEIGHT + (GAP_SIZE*2) + 15 + 25)
    end,

    ["SetTypeText"] = function(self, typeText)
        if typeText then
            self.textType:SetText("<"..typeText..">")
            self.textType:Show()
        else
            self.textType:SetText("")
            self.textType:Hide()
        end
    end,

    ["SetSelectedState"] = function(self, state)
        if state then
            if state == STATE_ID_LIST.ACCEPTED then
                self.acceptImageSmall:SetAlpha(1.0)
                self.declineImageSmall:SetAlpha(UNSELECTED_ALPHA)
            elseif state == STATE_ID_LIST.DECLINED then
                self.acceptImageSmall:SetAlpha(UNSELECTED_ALPHA)
                self.declineImageSmall:SetAlpha(1.0)
            else
                return self:SetSelectedState()
            end
        else
            self.acceptImageSmall:SetAlpha(UNSELECTED_ALPHA)
            self.declineImageSmall:SetAlpha(UNSELECTED_ALPHA)
        end
        self.selectedType = state
    end,

    ["SetSessionID"] = function(self, sessionID, sessionType)
        self.sessionID = sessionID
        self.sessionType = sessionType
    end,

    ["RefreshInviteTemplate"] = function(self)
        local inviteTemplate = self.inviteTemplate
        if not inviteTemplate then return end

        self:SetName(inviteTemplate:GetHost().displayName)
        self:SetDescription(inviteTemplate:GetDescription())
        self:SetTypeText(inviteTemplate:GetSessionID())
        self:SetSessionID(inviteTemplate:GetSessionID(), inviteTemplate:GetSessionType())
        self:SetSelectedState(inviteTemplate:GetOwnInviteState())
    end,

    ["RegisterInviteTemplateEvents"] = function(self)
        if not self.inviteTemplate then return end

        self.inviteTemplate:RegisterCallbackEvent("SessionTemplate_OnDescriptionChange", self.__cbOnDescriptionChange)
        self.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_PlayerInviteStateUpdate", self.__onPlayerInviteStateUpdate)
    end,

    ["UnregisterInviteTemplateEvents"] = function(self)
        if not self.inviteTemplate then return end

        self.inviteTemplate:UnregisterCallbackEvent("SessionTemplate_OnDescriptionChange", self.__cbOnDescriptionChange)
        self.inviteTemplate:UnregisterCallbackEvent("SessionTemplate_Invite_PlayerInviteStateUpdate", self.__onPlayerInviteStateUpdate)
    end,

    ["SetInviteTemplate"] = function(self, inviteTemplate)
        self:UnregisterInviteTemplateEvents()
        self.inviteTemplate = inviteTemplate
        self:RegisterInviteTemplateEvents()
        self:RefreshInviteTemplate()
    end,

    ["Disable"] = function(self, state)
        if state then
            self:SetSelectedState()
            self.frame:SetAlpha(UNSELECTED_ALPHA)
            self.isDisabled = true
        else
            self.frame:SetAlpha(1.0)
            self.isDisabled = nil
        end
    end
}


--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]
local function ButtonSetText(self, text)
    self.text:SetText(text)
    self:SetWidth(self.text:GetStringWidth() + 30)
end

local function NewButton(widget, content, buttonType)
    local button = CreateFrame("Button", nil, content, "UIPanelButtonTemplate")
    button:EnableMouse(true)
	button:SetScript("OnClick", AcceptDecline_OnClick)
    button.type = buttonType
    button.obj = widget

    local text = button:GetFontString()
	text:ClearAllPoints()
	text:SetPoint("TOPLEFT", 15, -1)
	text:SetPoint("BOTTOMRIGHT", -15, 1)
    text:SetJustifyV("MIDDLE")

    button._SetText = button.SetText
    button.SetText = ButtonSetText
    button.text = text

    return button
end

local function Constructor()
    local CreateLine = GGL.GUI.CreateLine

    local widget = {}
    widget.type  = Type

    local frame = GGL.GUI:CreateBorderedFrame(UIParent)
    widget.frame = frame

    local content = CreateFrame("Frame", nil, frame)
    content:SetPoint("TOPLEFT", frame, 5, -5)
    content:SetPoint("BOTTOMRIGHT", frame, -5, 5)
    content.obj = widget
    widget.content = content

    local textName = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlight")
    textName:SetPoint("TOPLEFT", content, GAP_SIZE+2, -(GAP_SIZE))
    textName:SetPoint("RIGHT", content, -GAP_SIZE, 0)
    textName:SetJustifyH("LEFT")
    textName:SetJustifyV("MIDDLE")
    textName:SetWordWrap(false)
    textName:SetHeight(TOP_HEIGHT)
    widget.textName = textName

    local declineImageSmall = content:CreateTexture(nil, "BACKGROUND")
    declineImageSmall:SetPoint("RIGHT", textName, 0, 0)
    declineImageSmall:SetWidth(TOP_HEIGHT-2)
    declineImageSmall:SetHeight(TOP_HEIGHT-2)
    declineImageSmall:SetAtlas(STATE_LIST[STATE_ID_LIST.DECLINED])
    widget.declineImageSmall = declineImageSmall

    local acceptImageSmall = content:CreateTexture(nil, "BACKGROUND")
    acceptImageSmall:SetPoint("RIGHT", declineImageSmall, "LEFT", -GAP_SIZE, 0)
    acceptImageSmall:SetWidth(TOP_HEIGHT-2)
    acceptImageSmall:SetHeight(TOP_HEIGHT-2)
    acceptImageSmall:SetAtlas(STATE_LIST[STATE_ID_LIST.ACCEPTED])
    widget.acceptImageSmall = acceptImageSmall

    local textType = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
    textType:SetPoint("RIGHT", acceptImageSmall, "LEFT", -15, 0)
    textType:SetPoint("LEFT", textName, 0, 0)
    textType:SetJustifyH("RIGHT")
    textType:SetJustifyV("MIDDLE")
    textType:SetWordWrap(false)
    textType:SetHeight(TOP_HEIGHT)
    textType:SetTextColor(0,1,0)
    textType:Hide()
    widget.textType = textType

    local line = content:CreateTexture(nil, "BACKGROUND")
    line:SetPoint("TOP", textName, "BOTTOM", 0, 0)
    line:SetPoint("LEFT", content, 0, 0)
    line:SetPoint("RIGHT", content, 0, 0)
    line:SetHeight(1)
    line:SetColorTexture(1, 1, 1, UNSELECTED_ALPHA)

    local textDesc = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
    textDesc:SetPoint("TOPLEFT", textName, "BOTTOMLEFT", 0, -(GAP_SIZE+1))
    textDesc:SetPoint("RIGHT", content, 0, 0)
    textDesc:SetJustifyH("LEFT")
    textDesc:SetJustifyV("TOP")
    textDesc:SetHeight(100)
    widget.textDesc = textDesc

    local acceptButton = NewButton(widget, content, STATE_ID_LIST.ACCEPTED)
    acceptButton:SetText(ACCEPT)
    acceptButton:SetPoint("TOPLEFT", textDesc, "BOTTOMLEFT", 0, -2)

    local declineButton = NewButton(widget, content, STATE_ID_LIST.DECLINED)
    declineButton:SetText(DECLINE)
    declineButton:SetPoint("LEFT", acceptButton, "RIGHT", 5, 0)

    -- methodes
    for method, func in pairs(methods) do
        widget[method] = func
    end

    -- cb
    widget.__cbOnDescriptionChange = function(inviteTemplate, description)
        widget:SetDescription(description)
    end

    widget.__onPlayerInviteStateUpdate = function(inviteTemplate, playerTemplate, newState)
        if playerTemplate ~= GGL.Templates.BNWoWPlayer:GetSelfPlayer() then return end
        widget:SetSelectedState(newState)
    end

    return AceGUI:RegisterAsWidget(widget)
end

AceGUI:RegisterWidgetType(Type, Constructor, Version)
