-- ###########################
-- EVENTS:
-- "Invite_NewInvite"      (inviteTemplate, description)
-- "Invite_InviteRemoved"  (inviteTemplate)
-- "Invite_InviteRequest"  (playerTemplate)
-- ###########################

local AddonName, GGL = ...
local L = GGL.Locale

local Invite = {}
GGL.Invite = Invite

-- callback handler
GGL:AddCallbackHandler(Invite)

-- ###########################
-- local
-- ###########################
-- lua
local setmetatable, pairs, ipairs = setmetatable, pairs, ipairs
local match, ssub = string.match, string.sub
local type = type
local time = time

-- WoW

-- other
local InviteSender, AnswerSender, DelistSender, StartSessionSender, RequestSender, UpdateSender    -- Sender for messsages
local Templates
local STATUS = GGL.INVITE_KEY_LIST

-- ###########################
-- Init
-- ###########################
local SenderList

local InviteTemplateSave = {}
local InvitationsSave = {}

local SESSION_ID_SPLIT, MSG_SPLIT, ID_SPLIT = "\001", "\002", "\003"
local SESSION_SEND, SESSION_RECEIVE = "%s"..SESSION_ID_SPLIT, "^(.*)"..SESSION_ID_SPLIT.."(.*)"
local ID_SPLIT_SEND, ID_SPLIT_RECEIVE = "%s"..ID_SPLIT, "^(.*)"..ID_SPLIT.."(.*)"
local FORMAT_STRING = {
    Invite = {
        --- 1 = send out invite     	//      sessionID, description
        --- 2 = send invite answer      //      sessionID, answerID
        --- 3 = update description      //      sessionID, description
        send = SESSION_SEND..ID_SPLIT_SEND.."%s",
        --receive = "(.*)",
    },
    SessionUpdate = {
        -- 1 = start: sessionID, duration, endTime, currentTime, settings
        send_start = SESSION_SEND.."1"..ID_SPLIT.."%d"..MSG_SPLIT.."%d"..MSG_SPLIT.."%d"..MSG_SPLIT.."%s",
        receive_start = "(%d*)"..MSG_SPLIT.."(%d*)"..MSG_SPLIT.."(%d*)"..MSG_SPLIT.."(.*)",
        -- 2 = end: sessionID
        send_end = SESSION_SEND.."2"..ID_SPLIT.."",
        -- 3 = delist: sessionID
        send_delist = SESSION_SEND.."3"..ID_SPLIT.."",
        -- 5 = config: sessionID, configString
        send_config = SESSION_SEND.."4"..ID_SPLIT.."%s",
        -- 5 = stop: sessionID
        send_stop = SESSION_SEND.."5"..ID_SPLIT.."",
    },
    PlayerUpdate = {
        -- displayName, currentServerTime, ggRewardID
        send = "%d"..MSG_SPLIT.."%s"..MSG_SPLIT.."%d",
        receive = "^(%d*)"..MSG_SPLIT.."(.*)"..MSG_SPLIT.."(%d)",
    },
    RequestInvite = {
        -- id, addonVersion
        send = "1"..ID_SPLIT.."%s",
        --receive = "",
        send_answer = "2"..ID_SPLIT.."%s",
        --receive_answer = "^(.*)"..MSG_SPLIT.."(.*)",
    }
}

local MessageReceiverFunctions = {
    Invite = function(inviteTemplate, senderPlayer, data)
        if inviteTemplate.isClosed then return end
        local id, idData = match(data, ID_SPLIT_RECEIVE)

        if id == "1" and not InvitationsSave[inviteTemplate] then   -- new invite
            InvitationsSave[inviteTemplate] = senderPlayer
            inviteTemplate:SetDescription(idData)
            Invite:FireCallbackEvent("Invite_NewInvite", inviteTemplate, idData)
            GGL:Print(format(L["New invite from %s."], senderPlayer.displayName))
        elseif id == "2" then   -- get answer for an invite
            -- only the host needs a answer for invites..
            if not inviteTemplate:PlayerIsHost() then return end
            inviteTemplate:SetPlayerInviteState(senderPlayer, idData)
        elseif id == "3" then   -- description update
            if not inviteTemplate:PlayerTemplateIsHost(senderPlayer) then return end
            inviteTemplate:SetDescription(idData)
        end
    end,

    SessionUpdate = function(inviteTemplate, senderPlayer, data)
        -- only host can update a session
        if not inviteTemplate:PlayerTemplateIsHost(senderPlayer) then return end

        local id, idData = match(data, ID_SPLIT_RECEIVE)

        if id == "1" then   -- start session
            -- cant start closed invites
            if inviteTemplate.isClosed then return end
            -- get data
            local duration, endTime, currentTime, settings = match(idData, FORMAT_STRING.SessionUpdate.receive_start)
            if not duration then return end

            senderPlayer:SetTimeOffset(GGL.GetTimeOffset(tonumber(currentTime)))
            -- send own player data
            Invite:SendPlayerData(inviteTemplate)

            inviteTemplate:HostSendStartSession(tonumber(duration), tonumber(endTime), settings)
        elseif id == "2" then   -- end session
            inviteTemplate:HostSendEndSession()
        elseif id == "3" then   -- delist session
            inviteTemplate:HostSendDelistSession()
        elseif id == "4" then   -- new settings
            if inviteTemplate.sessionIsRunning then
                inviteTemplate.session:SetSettingsString(idData)
            end
        elseif id == "5" then   -- stop session
            inviteTemplate:HostSendStopSession()
        end

        -- this fires when session ends and when session is removed
        if id == "2" or id == "3" or id == "5" then   -- session removed
            InvitationsSave[inviteTemplate] = nil
            Invite:FireCallbackEvent("Invite_InviteRemoved", inviteTemplate)
        end
    end,
}

local function OnSenderMessageReceive(self, prefix, sender, message, isBNMessage)
    if not message or message == "" then return end

    -- get sessionID and data
    local sessionID, data = match(message, SESSION_RECEIVE)
    if not sessionID then return end

    -- get playerTemplate
    local senderPlayer
    if isBNMessage then
        senderPlayer = Templates.BNWoWPlayer:GetByGameAccountID(sender)
    else
        senderPlayer = Templates:Get("WoWPlayer", sender)
    end
    if not senderPlayer then return end

    -- get inviteHandler
    local inviteTemplate
    if InviteTemplateSave[sessionID] then
        inviteTemplate = InviteTemplateSave[sessionID]
    else
        inviteTemplate = Templates.Invite:Get(sessionID, senderPlayer)
        if not inviteTemplate then return end
        Invite:RegisterInviteTemplate(inviteTemplate)
    end

    -- handle the prefix
    if MessageReceiverFunctions[prefix] then
        MessageReceiverFunctions[prefix](inviteTemplate, senderPlayer, data)
    end
end

local function OnSenderMessageReceive_RequestInvite(self, prefix, sender, message, isBNMessage)
    if not message or message == "" then return end

    -- get playerTemplate
    local senderPlayer
    if isBNMessage then
        senderPlayer = Templates.BNWoWPlayer:GetByGameAccountID(sender)
    else
        senderPlayer = Templates:Get("WoWPlayer", sender)
    end
    if not senderPlayer then return end

    -- set player online
    if senderPlayer.SetOnline then senderPlayer:SetOnline(true) end

    local id, idData = match(message, ID_SPLIT_RECEIVE)

    senderPlayer:SetGGLData(idData or "unknown")

    if id == "1" then   --  get a request
        senderPlayer:SetGGLData(idData)

        -- the player requests invites so we should remove open invites as
        -- old invites maybe no longer aviable
        for inviteTemplate, hostPlayerTemplate in pairs(InvitationsSave) do
            if hostPlayerTemplate == senderPlayer and not inviteTemplate.sessionIsRunning then
                InvitationsSave[inviteTemplate] = nil
                Invite:FireCallbackEvent("Invite_InviteRemoved", inviteTemplate)
            end
        end

        -- send addon data
        SenderList.RequestInvite:SendMessage(senderPlayer, format(FORMAT_STRING.RequestInvite.send_answer, GGL.ADDON_VERSION), true)

        -- send a playerUpdate
        -- Invite:SendPlayerDataToPlayer(senderPlayer)

        Invite:FireCallbackEvent("Invite_InviteRequest", senderPlayer)

        -- send out invites for registered sessions
        for sessionID, invTemplate in pairs(InviteTemplateSave) do
            invTemplate:OnPlayerRequestInvite(senderPlayer)
        end
    elseif id == "2" then   -- get a answer
        -- set players used version
        senderPlayer:SetGGLData(idData)
    end
end

local function OnSenderMessageReceive_PlayerUpdate(self, prefix, sender, message, isBNMessage)
    if not message or message == "" then return end

    -- get playerTemplate
    local senderPlayer
    if isBNMessage then
        senderPlayer = Templates.BNWoWPlayer:GetByGameAccountID(sender)
    else
        senderPlayer = Templates:Get("WoWPlayer", sender)
    end
    if not senderPlayer then return end

    -- set player online
    if senderPlayer.SetOnline then senderPlayer:SetOnline(true) end

    -- check for ggl data
    if not senderPlayer.gglVersion then
        -- no ggl data, but we get player data?
        senderPlayer:SetGGLData("unknown")
        SenderList.RequestInvite:SendMessage(senderPlayer, format(FORMAT_STRING.RequestInvite.send_answer, GGL.ADDON_VERSION), true)
    end

    -- set data
    local currentServerTime, displayName, ggRewardID = match(message, FORMAT_STRING.PlayerUpdate.receive)
    senderPlayer:SetDisplayName(displayName)
    senderPlayer:SetGGRewardID(ggRewardID)
    senderPlayer:SetTimeOffset(GGL.GetTimeOffset(tonumber(currentServerTime)))
end

local function PlayerEvent_NewPlayerRegistered(player)
    -- send out a request for invites
    -- this also fires when we login/reload so it should "reload" invites also
    Invite:RequestInvites(player)
end

local function PlayerEvent_DisplayNameUpdate(player)
    -- only send out new player data when own data gets refreshed
    -- otherwise this gonna be an endless loop
    if player and player.isPlayer then
        for inviteTemplate in pairs(InvitationsSave) do
            Invite:SendPlayerData(inviteTemplate)
        end
    end
end

local function Init()

    SenderList = {
        Invite          = GGL.Sender:RegisterReceiver("Invite", OnSenderMessageReceive),
        SessionUpdate   = GGL.Sender:RegisterReceiver("SessionUpdate", OnSenderMessageReceive),
        PlayerUpdate    = GGL.Sender:RegisterReceiver("PlayerUpdate", OnSenderMessageReceive_PlayerUpdate),
        RequestInvite   = GGL.Sender:RegisterReceiver("RequestInvite", OnSenderMessageReceive_RequestInvite, true),
    }

    Templates = GGL.Templates

    -- Player data
    Templates.WoWPlayer:RegisterCallbackEvent("PlayerTemplate_NewPlayerRegistered", PlayerEvent_NewPlayerRegistered)
    Templates.WoWPlayer:RegisterCallbackEvent("PlayerTemplate_DisplayNameUpdate", PlayerEvent_DisplayNameUpdate)
    Templates.BNWoWPlayer:RegisterCallbackEvent("PlayerTemplate_NewPlayerRegistered", PlayerEvent_NewPlayerRegistered)
    Templates.BNWoWPlayer:RegisterCallbackEvent("PlayerTemplate_DisplayNameUpdate", PlayerEvent_DisplayNameUpdate)
    Templates.BNWoWPlayer:RegisterCallbackEvent("PlayerTemplate_RewardIDUpdate", PlayerEvent_DisplayNameUpdate) -- use displayName updateID
end
GGL:AddInitFuncPlayerLogin(Init)


-- ###########################
-- Invite functions
-- ###########################

--- Sends own playerdata to a player
-- @param       targetPlayerTemplate      <playerTemplate>
function Invite:SendPlayerDataToPlayer(targetPlayerTemplate)
    if not targetPlayerTemplate then return end

    SenderList.PlayerUpdate:SendMessage(targetPlayerTemplate,
        format(FORMAT_STRING.PlayerUpdate.send,
            GGL.GetTime(),
            GGL.Templates.BNWoWPlayer:GetSelfPlayer():GetDisplayName(),
            GGL.Templates.BNWoWPlayer:GetSelfPlayer():GetGGRewardID()
        ),
        true
    )
end

--- Requests invite from players in battle.net list
-- @para        playerTemplate      <playerTemplate/bool> If true then invites get send out to whole friendlist
function Invite:RequestInvites(playerTemplate)
    if playerTemplate == true then
        for playerGUID, player in GGL.Templates.BNWoWPlayer:IterPlayerList() do
            Invite:RequestInvites(player)
        end
    elseif playerTemplate then
        SenderList.RequestInvite:SendMessage(playerTemplate, format(FORMAT_STRING.RequestInvite.send, GGL.ADDON_VERSION), true)
    end
end

--- Registers a new inviteTemplate
--- Without register the receiver cant fire a event
-- @param       inviteTemplate      <inviteTemplate>
function Invite:RegisterInviteTemplate(inviteTemplate)
    InviteTemplateSave[inviteTemplate:GetSessionID()] = inviteTemplate
end

--- Unregister a inviteTemplate
-- @param       inviteTemplate      <inviteTemplate>
function Invite:UnregisterInviteTemplate(inviteTemplate)
    --disabled for now
    --InviteTemplateSave[inviteTemplate:GetSessionID()] = nil
end

--- Returns table with inviteTemplates
-- @format      [inviteTemplate] = senderPlayerTemplate
function Invite:GetInvitationList()
    return InvitationsSave
end

--- Sends out a player invite for a inviteTemplate
-- @param       inviteTemplate      <inviteTemplate>
-- @param       playerTemplate      <playerTemplate>
-- @param       sendInstant         <bool> send out invite without delay
function Invite:SendInvite(inviteTemplate, playerTemplate, sendInstant)
    if not inviteTemplate:PlayerIsHost() then return end

    SenderList.Invite:SendMessage(playerTemplate,
        format(FORMAT_STRING.Invite.send,
            inviteTemplate:GetSessionID(),
            1,
            inviteTemplate:GetDescription()
        ),
        sendInstant
    )
end

--- Sends out a player invite for a inviteTemplate
-- @param       inviteTemplate      <inviteTemplate>
-- @param       playerTemplate      <playerTemplate>
function Invite:SendInviteState(inviteTemplate, playerTemplate)
    if inviteTemplate:PlayerIsHost() then return end

    SenderList.Invite:SendMessage(playerTemplate,
        format(FORMAT_STRING.Invite.send,
            inviteTemplate:GetSessionID(),
            2,
            inviteTemplate:GetOwnInviteState()
        )
    )
end

--- Sends out new description for a inviteTemplate
-- @param       inviteTemplate      <inviteTemplate>
-- @param       playerTemplate      <playerTemplate>
function Invite:SendDescription(inviteTemplate, playerTemplate)
    if not inviteTemplate:PlayerIsHost() then return end

    SenderList.Invite:SendMessage(playerTemplate,
        format(FORMAT_STRING.Invite.send,
            inviteTemplate:GetSessionID(),
            3,
            inviteTemplate:GetDescription()
        )
    )
end

--- Sends out new set config for a inviteTemplate
-- @param       inviteTemplate      <inviteTemplate>
-- @param       playerTemplate      <playerTemplate>
function Invite:SendSettings(inviteTemplate, playerTemplate)
    if not inviteTemplate:PlayerIsHost() then return end

    SenderList.SessionUpdate:SendMessage(playerTemplate,
        format(FORMAT_STRING.SessionUpdate.send_config,
            inviteTemplate:GetSessionID(),
            inviteTemplate:GetSettings()
        )
    )
end

--- Sends out the session start for a inviteTemplate
-- @param       inviteTemplate      <inviteTemplate>
-- @param       playerTemplate      <playerTemplate>
function Invite:SendStart(inviteTemplate, playerTemplate)
    if not inviteTemplate:PlayerIsHost() then return end
    SenderList.SessionUpdate:SendMessage(playerTemplate,
        format(FORMAT_STRING.SessionUpdate.send_start,
            inviteTemplate:GetSessionID(),
            inviteTemplate:GetDuration(),
            inviteTemplate:GetEndTime(),
            GGL.GetTime(),
            inviteTemplate:GetSettings()
        )
    )
end

--- Sends out the session stop for a inviteTemplate
-- @param       inviteTemplate      <inviteTemplate>
-- @param       playerTemplate      <playerTemplate>
function Invite:SendStop(inviteTemplate, playerTemplate)
    if not inviteTemplate:PlayerIsHost() then return end

    SenderList.SessionUpdate:SendMessage(playerTemplate,
        format(FORMAT_STRING.SessionUpdate.send_stop,
            inviteTemplate:GetSessionID()
        )
    )
end

--- Sends out the session end for a inviteTemplate
-- @param       inviteTemplate      <inviteTemplate>
-- @param       playerTemplate      <playerTemplate>
function Invite:SendEnd(inviteTemplate, playerTemplate)
    if not inviteTemplate:PlayerIsHost() then return end

    SenderList.SessionUpdate:SendMessage(playerTemplate,
        format(FORMAT_STRING.SessionUpdate.send_end,
            inviteTemplate:GetSessionID()
        )
    )
end

--- Sends out the session delist for a inviteTemplate
-- @param       inviteTemplate      <inviteTemplate>
-- @param       playerTemplate      <playerTemplate>
function Invite:SendDelist(inviteTemplate, playerTemplate)
    if not inviteTemplate:PlayerIsHost() then return end

    SenderList.SessionUpdate:SendMessage(playerTemplate,
        format(FORMAT_STRING.SessionUpdate.send_delist,
            inviteTemplate:GetSessionID()
        )
    )
end

--- Sends own playerdata to the host
-- @param       inviteTemplate      <inviteTemplate>
function Invite:SendPlayerData(inviteTemplate)
    if inviteTemplate:PlayerIsHost() then return end

    self:SendPlayerDataToPlayer(inviteTemplate:GetHost())
end